#include <stdlib.h>
#include <stdio.h>

#include "btree.h"

/*
 * Realiza a busca binaria na pagina
 * Devolve a posicao da chave encontrada ou posicao do ponteiro para a
 *  pagina filha */
int busca_binaria(Pagina *no,  int info) 
{ 
	int meio, i, f; 

	i = 0;
	f = no->num_chaves-1;

	while (i <= f)
	{
		meio = (i + f)/2;
		if (no->chaves[meio] == info)
			return(meio);	// Encontrou, devolve a posição em que a chave esta
		else if (no->chaves[meio] > info)
			f = meio - 1;
		else i = meio + 1;
	}
	return(i); //Nao encontrou, devolve a posicao do ponteiro para o filho
}

/*
 * Realiza a busca por um determinado valor
 * Devolve TRUE se entrar, FALSE caso contrario */
bool busca(Pagina *raiz, int info)
{
	Pagina *no;
	int pos; //posicao devolvida pela busca binaria.

	no = raiz;
	while (no != NULL)
	{
		pos = busca_binaria(no, info);
		if (pos < no->num_chaves && no->chaves[pos] == info)
			return(true);
		else no = no->filhos[pos];
	}
	return(false);
}

void imprime_chave(int chave)
{
    printf("%d\n",chave);
}

/*
 * Percorre a arvore em ordem */
void em_ordem(Pagina *raiz)
{
	int i;

	if (raiz != NULL)
	{
		for (i = 0; i < raiz->num_chaves; i++)
		{
			em_ordem(raiz->filhos[i]);
			imprime_chave(raiz->chaves[i]);
		}
		em_ordem(raiz->filhos[i]);
	}
}

/*
 * Insere uma chave e o ponteiro para o filho da direita em um no */
void insere_chave(Pagina *raiz, int info, Pagina *filhodir)
{
	int k, pos;

	// busca para obter a posicao ideal para inserir a nova chave
	pos = busca_binaria(raiz, info);
	k = raiz->num_chaves;

	// realiza o remanejamento para manter as chaves ordenadas
	while (k > pos && info < raiz->chaves[k-1])
	{
		raiz->chaves[k] = raiz->chaves[k-1];
		raiz->filhos[k+1] = raiz->filhos[k];
		k--;
	}
	// insere a chave na posicao ideal
	raiz->chaves[pos] = info;
	raiz->filhos[pos+1] = filhodir;
	raiz->num_chaves++;
}

/* 
 * Realiza a busca do no para inserir a chave e faz as subdivisoes quando necessarias */
Pagina *insere(Pagina *raiz, int info, bool *h, int *info_retorno)
{
	int i, j, pos;
	int info_mediano; 	//auxiliar para armazenar a chave que ira subir para o pai
	Pagina *temp, *filho_dir;	//ponteiro para o filho a direita da chave

	if (raiz == NULL)
	{
		//O no atual e o ideal para inserir a nova chave (chegou em um no folha)
		*h = true;
		*info_retorno = info;
		return(NULL);
	}
	else {
		pos = busca_binaria(raiz,info);
		if (raiz->num_chaves > pos && raiz->chaves[pos] == info)
		{
			//printf("Chave já contida na Árvore");
			*h = false;
		}
		else {
			//desce na arvore ate encontrar o no folha para inserir a chave.
			filho_dir = insere(raiz->filhos[pos],info,h,info_retorno);
			if (*h) //Se TRUE deve inserir a info_retorno no no.
			{
				if (raiz->num_chaves < MAX_CHAVES)	// Tem espaço na pagina.
				{
					insere_chave(raiz, *info_retorno, filho_dir);
					*h = false;
				}
				else { //Overflow. Precisa subdividir.
					temp = (Pagina *) malloc (sizeof(Pagina));
					temp->num_chaves = 0;

					// Inicializa filhos com NULL
					for (i = 0; i < MAX_FILHOS; i++)
						temp->filhos[i] = NULL;

					// Elemento mediano que vai subir para o pai
					info_mediano = raiz->chaves[MIN_OCUP];

					// Insere metade do no raiz no temp (efetua subdivisao)
					temp->filhos[0] = raiz->filhos[MIN_OCUP+1];
					for (i = MIN_OCUP + 1; i < MAX_CHAVES; i++)
						insere_chave(temp, raiz->chaves[i], raiz->filhos[i+1]);

					// Atualiza no raiz. 
					for (i = MIN_OCUP; i<MAX_CHAVES; i++)
					{
						raiz->chaves[i] = 0;
						raiz->filhos[i+1] = NULL;
					}
					raiz->num_chaves = MIN_OCUP;

					// Verifica em qual no sera inserida a nova chave
					if (pos <= MIN_OCUP)
						insere_chave(raiz, *info_retorno, filho_dir);
					else insere_chave(temp, *info_retorno, filho_dir);

					// Devolve o mediano para inseri-lo no pai e o temp como filho direito do mediano.
					*info_retorno = info_mediano;
					return(temp);
				}
			}
		}
	}
}

Pagina *insere_arvoreB(Pagina *raiz, int info)
{
	bool h;
	int info_retorno, i;
	Pagina *filho_dir, *nova_raiz;

	filho_dir = insere(raiz,info,&h,&info_retorno);
	if (h)
	{ //Aumetara a altura da árvore
		nova_raiz = (Pagina *) malloc(sizeof(Pagina));
		nova_raiz->num_chaves = 1;
		nova_raiz->chaves[0] = info_retorno;
		nova_raiz->filhos[0] = raiz;
		nova_raiz->filhos[1] = filho_dir;
		for (i = 2; i <= MAX_CHAVES; i++)
			nova_raiz->filhos[i] = NULL;
		return(nova_raiz);
	}
	else return(raiz);
}
